import os

from ..config import settings
from ..common import progress
from ..common.tools import cprint
from kiutils.symbol import SymbolLib


# KiCad Component Library Manager
class ComponentLibManager(object):
    def __init__(self, library_path):
        # Load library and template paths
        cprint(f'[KCAD]\tlibrary_path: {library_path}', silent=settings.SILENT)

        # Check files exist
        if not os.path.isfile(library_path):
            cprint(f'[KCAD]\tError loading library file ({library_path})', silent=settings.SILENT)
            return None

        # Load library
        self.kicad_lib = SymbolLib.from_file(library_path)
        self.library_name = library_path.split(os.sep)[-1]
        cprint('[KCAD]\tNumber of parts in library ' + self.library_name + ': ' + str(len(self.kicad_lib.symbols)), silent=settings.SILENT)

    def is_symbol_in_library(self, symbol_id):
        ''' Check if symbol already exists in library '''
        for symbol in self.kicad_lib.symbols:
            cprint(f'[DBUG]\t{symbol.libId} ?= {symbol_id}', silent=settings.HIDE_DEBUG)
            if symbol.libId == symbol_id:
                cprint(f'[KCAD]\tWarning: Component {symbol_id} already in library', silent=settings.SILENT)
                return True

        return False

    def add_symbol_to_library_from_inventree(self, symbol_data, template_path=None, show_progress=True):
        ''' Create symbol in KiCad library '''
        part_in_lib = False
        new_part = False
        part_name = ''
        parameters = symbol_data.get('parameters', {})
        parameters = {**symbol_data, **parameters}
        key_list = list(parameters.keys())
        key_list.sort(key=len, reverse=True)

        def replace_wildcards(field):
            for key in key_list:
                if key in field:
                    field = field.replace(key, parameters[key])
            return field

        symbol_id = symbol_data.get('Symbol', '').split(':')
        if not symbol_id:
            cprint('[KCAD] Error: Adding a new symbol to a KiCad library requires the \'Symbol\' key with the following format: {lib}:{symbol_id}')
            return part_in_lib, new_part, part_name

        if not template_path:
            category = symbol_data['Template'][0]
            subcategory = symbol_data['Template'][1]

            # Fetch template path
            try:
                template_path = settings.symbol_templates_paths[category][subcategory]
            except:
                template_path = settings.symbol_templates_paths[category]['Default']

        # Check files exist
        if not self.kicad_lib:
            return part_in_lib, new_part
        if not os.path.isfile(template_path):
            cprint(f'[KCAD]\tError loading template file ({template_path})', silent=settings.SILENT)
            return part_in_lib, new_part, part_name

        # Load template
        templatelib = SymbolLib.from_file(template_path)
        # Load new symbol
        if len(templatelib.symbols) == 1:
            for symbol in templatelib.symbols:
                new_symbol = symbol
        else:
            cprint('[KCAD]\tError: Found more than 1 symbol template in template file, aborting', silent=settings.SILENT)
            return part_in_lib, new_part, part_name

        # Update name/ID
        part_name = replace_wildcards(new_symbol.libId)
        new_symbol.libId = part_name

        # Check if part already in library
        try:
            is_symbol_in_library = self.is_symbol_in_library(part_name)
            part_in_lib = True
        except:
            is_symbol_in_library = False
        if is_symbol_in_library:
            return part_in_lib, new_part, part_name

        # Progress Update
        if not progress.update_progress_bar(show_progress):
            return part_in_lib, new_part, part_name

        # Update properties
        for property in new_symbol.properties:
            property.value = replace_wildcards(property.value)

        # Add symbol to library
        self.kicad_lib.symbols.append(new_symbol)
        # Write library
        self.kicad_lib.to_file()

        cprint(f'[KCAD]\tSuccess: Component added to library {self.library_name}', silent=settings.SILENT)
        part_in_lib = True
        new_part = True

        # Progress Update
        if not progress.update_progress_bar(show_progress):
            pass

        return part_in_lib, new_part, part_name



from ..config import settings
from ..common.tools import download, download_with_retry

API_BASE_URL = 'https://snapeda.eeinte.ch/?'
SNAPEDA_URL = 'https://www.snapeda.com'


def fetch_snapeda_part_info(part_number: str) -> dict:
    ''' Fetch SnapEDA part data from API '''

    api_url = API_BASE_URL + part_number.replace(' ', '%20')
    data = download(api_url, timeout=10)
    return data if data else {}


def parse_snapeda_response(response: dict) -> dict:
    ''' Return only relevant information from SnapEDA API response '''

    data = {
        'part_number': None,
        'has_symbol': False,
        'has_footprint': False,
        'symbol_image': None,
        'footprint_image': None,
        'package': None,
        'part_url': None,
        'has_single_result': False,
    }

    number_results = int(response.get('hits', 0))

    # Check for single result
    if number_results == 1:
        try:
            data['part_number'] = response['results'][0].get('part_number', None)
            data['has_symbol'] = response['results'][0].get('has_symbol', False)
            data['has_footprint'] = response['results'][0].get('has_footprint', False)
            data['package'] = response['results'][0]['package'].get('name', None)
            data['part_url'] = SNAPEDA_URL + response['results'][0]['_links']['self'].get('href', '')
            data['part_url'] += '?ref=kintree'
            data['has_single_result'] = True
        except KeyError:
            pass

        # Separate as the 'models' key does not always exist
        try:
            data['symbol_image'] = response['results'][0]['models'][0]['symbol_medium'].get('url', None)
        except KeyError:
            pass
        try:
            data['footprint_image'] = response['results'][0]['models'][0]['package_medium'].get('url', None)
        except KeyError:
            pass
    elif number_results > 1:
        try:
            data['part_url'] = SNAPEDA_URL + '/search/' + response['pages'][0].get('link', None).split('&')[0] + '&ref=kintree'
        except:
            pass
    else:
        pass

    return data


def download_snapeda_images(snapeda_data: dict) -> dict:
    ''' Download symbol and footprint images from SnapEDA's server '''

    images = {
        'symbol': None,
        'footprint': None,
    }

    try:
        part_number = snapeda_data["part_number"].replace('/', '').lower()
    except:
        part_number = None

    if part_number:
        try:
            if snapeda_data['symbol_image']:
                # Form path
                image_name = f'{part_number}_symbol.png'
                image_location = settings.search_images + image_name

                # Download symbol image
                symbol = download_with_retry(snapeda_data['symbol_image'], image_location, filetype='Image')
                if symbol:
                    images['symbol'] = image_location
        except KeyError:
            pass

        try:
            if snapeda_data['footprint_image']:
                # Form path
                image_name = f'{part_number}_footprint.png'
                image_location = settings.search_images + image_name

                # Download symbol image
                footprint = download_with_retry(snapeda_data['footprint_image'], image_location, filetype='Image')
                if footprint:
                    images['footprint'] = image_location
        except KeyError:
            pass

    return images


def test_snapeda_api() -> bool:
    ''' Test method for SnapEDA API '''

    result = False

    # Test single result
    response = fetch_snapeda_part_info('NTR06B2001CTRF')
    data = parse_snapeda_response(response)
    images = download_snapeda_images(data)

    if data['part_number'] and data['has_symbol'] and images['symbol']:
        result = True

    # Test multiple results
    if result:
        response = fetch_snapeda_part_info('1N4148W-7-F')
        data = parse_snapeda_response(response)
        if data['has_single_result']:
            result = False

    return result



from enum import Enum
from math import pi
from typing import Optional, List

import flet as ft

GUI_PARAMS = {
    'nav_rail_min_width': 100,
    'nav_rail_width': 400,
    'nav_rail_alignment': -0.9,
    'nav_rail_icon_size': 40,
    'nav_rail_text_size': 16,
    'nav_rail_padding': 10,
    'textfield_width': 600,
    'textfield_dense': True,
    'textfield_space_after': 3,
    'dropdown_width': 600,
    'dropdown_dense': False,
    'searchfield_width': 300,
    'button_width': 110,
    'button_height': 56,
    'icon_size': 40,
    'text_size': 16,
}
# Contains data from all views
data_from_views = {}


class DialogType(Enum):
    VALID = 'valid'
    WARNING = 'warning'
    ERROR = 'error'


def handle_transition(page: ft.Page, transition: bool, update_page=False, timeout=0):
    # print(f'{transition=} | {update_page=} | {timeout=}')
    if transition:
        transition = ft.PageTransitionTheme.CUPERTINO
        page.theme.page_transitions.android = transition
        page.theme.page_transitions.ios = transition
        page.theme.page_transitions.linux = transition
        page.theme.page_transitions.macos = transition
        page.theme.page_transitions.windows = transition
    else:
        page.theme.page_transitions.android = ft.PageTransitionTheme.NONE
        page.theme.page_transitions.ios = ft.PageTransitionTheme.NONE
        page.theme.page_transitions.linux = ft.PageTransitionTheme.NONE
        page.theme.page_transitions.macos = ft.PageTransitionTheme.NONE
        page.theme.page_transitions.windows = ft.PageTransitionTheme.NONE

    # Wait
    if timeout:
        import time
        time.sleep(timeout)

    # Update
    if update_page:
        page.update()


def update_theme(page: ft.Page, mode='light', transition=False, compact=True):
    # Color theme
    page.theme_mode = mode

    # UI theme
    theme = ft.Theme()
    page.theme = theme

    # Make it more compact
    if compact:
        page.theme.visual_density = ft.ThemeVisualDensity.COMPACT
    else:
        page.theme.visual_density = ft.ThemeVisualDensity.STANDARD

    # Disable transitions by default
    handle_transition(page, transition=False)


class CommonView(ft.View):
    '''Common view to all GUI views'''

    _page = None
    navigation_rail = None
    title = None
    column = None
    fields = None
    data = None
    dialog = None
    
    def __init__(self, page: ft.Page, appbar: ft.AppBar, navigation_rail: ft.NavigationRail):
        # Store page pointer
        self._page = page

        # Init view
        super().__init__(route=self.route, appbar=appbar)

        # Set navigation rail
        if not self.navigation_rail:
            self.navigation_rail = navigation_rail

    def build_column(self):
        # Empty column (to be set inside the children views)
        self.column = ft.Column()

    def build(self):
        # Build column
        if not self.column:
            self.build_column()
        # Set view controls
        self.controls = [
            ft.Row(
                controls=[
                    self.navigation_rail,
                    ft.VerticalDivider(width=1),
                    self.column,
                ],
                expand=True,
            ),
        ]

    def build_dialog(self):
        return None
    
    def build_snackbar(self, d_type: DialogType, message: str):
        if d_type == DialogType.VALID:
            self.dialog = ft.SnackBar(
                bgcolor=ft.colors.GREEN_100,
                content=ft.Text(
                    message,
                    color=ft.colors.GREEN_700,
                    size=GUI_PARAMS['nav_rail_text_size'],
                    weight=ft.FontWeight.BOLD,
                ),
            )
        elif d_type == DialogType.WARNING:
            self.dialog = ft.SnackBar(
                bgcolor=ft.colors.AMBER_100,
                content=ft.Text(
                    message,
                    color=ft.colors.AMBER_800,
                    size=GUI_PARAMS['nav_rail_text_size'],
                    weight=ft.FontWeight.BOLD,
                ),
            )
        elif d_type == DialogType.ERROR:
            self.dialog = ft.SnackBar(
                bgcolor=ft.colors.RED_100,
                content=ft.Text(
                    message,
                    color=ft.colors.RED_700,
                    size=GUI_PARAMS['nav_rail_text_size'],
                    weight=ft.FontWeight.BOLD,
                ),
            )

    def show_dialog(
            self,
            d_type: Optional[DialogType] = None,
            message: Optional[str] = None,
            snackbar=True,
            open=True,
    ):
        if snackbar:
            self.build_snackbar(d_type, message)
        if isinstance(self.dialog, ft.SnackBar):
            self._page.snack_bar = self.dialog
            self._page.snack_bar.open = True
        elif isinstance(self.dialog, ft.Banner):
            self._page.banner = self.dialog
            self._page.banner.open = open
        elif isinstance(self.dialog, ft.AlertDialog):
            self._page.dialog = self.dialog
            self._page.dialog.open = open
        self._page.update()


class SwitchWithRefs(ft.Switch):
    '''Link the visibility of other fields to a switch value'''

    linked_refs = []
    
    def __init__(
        self,
        refs: List[ft.Ref] = None,
        reverse_dir: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if refs:
            self.refs = refs
            self.enable_refs(self.value)
        self.reverse_dir = reverse_dir

    def enable_refs(self, enable):
        if self.reverse_dir:
            enable = not enable
        for ref in self.linked_refs:
            ref.current.visible = enable
            try:
                ref.current.update()
            except AssertionError:
                # Control not added to page yet
                pass
    
    def process_change(self, e, handler, *args, **kwargs):
        enable = False
        if e.data == 'true':
            enable = True
        self.enable_refs(enable)
        handler(e, *args, **kwargs)

    @property
    def refs(self):
        return self.linked_refs
    
    @refs.setter
    def refs(self, references: List[ft.Ref]):
        if references:
            self.linked_refs = []
            for ref in references:
                try:
                    if ref.current is None:
                        raise Exception(f'Reference "{ref.current}" needs to be added to the page first')
                except AttributeError:
                    raise Exception(f'"{ref}" is not a Flet Ref (type: {type(ref)})')
                # if ft.Control not in ref.current.__class__.__mro__:
                #     raise Exception(f'"{ref.current}" is not a Flet Control ({type(ref.current)})')
                self.linked_refs.append(ref)
            if self.linked_refs:
                self.enable_refs(self.value)

    @ft.Switch.on_change.setter
    def on_change(self, handler, *args, **kwargs):
        ft.Switch.on_change.fset(
            self,
            lambda e: self.process_change(e, handler, *args, **kwargs)
        )


class DropdownWithSearch(ft.UserControl):
    '''Implements a dropdown with search box'''

    dropdown = None
    search_button = None
    search_field = None
    search_box = None
    search_width = None
    
    def build(self):
        return ft.Row([
            self.dropdown,
            self.search_box,
            self.search_button,
        ])
    
    def __str__(self):
        return f'dropdown_with_search {{dropdown: {self.dropdown}, search_field: {self.search_field}}}'

    def __init__(
        self,
        label: Optional[str] = None,
        dr_width: Optional[int] = None,
        sr_width: Optional[int] = None,
        dense: Optional[bool] = None,
        disabled=False,
        sr_animate=100,
        options=None,
        on_change=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._options = options
        self.dropdown = ft.Dropdown(
            label=label,
            width=dr_width,
            dense=dense,
            options=options,
            on_change=on_change,
        )
        self.search_button = ft.IconButton(
            'search',
            on_click=self.search_now
        )
        self.search_field = ft.TextField(
            border="none",
            width=sr_width,
            dense=dense,
            on_change=self.on_search,
        )
        self.search_box = ft.Container(
            content=self.search_field,
            width=0,
            animate=ft.Animation(sr_animate),
        )
        self.disabled = disabled
        self.search_width = sr_width

    @property
    def label(self):
        return self.dropdown.label
    
    @label.setter
    def label(self, label):
        self.dropdown.label = label
        
    @property
    def value(self):
        return self.dropdown.value
    
    @value.setter
    def value(self, value):
        self.dropdown.value = value
        if value is None:
            self.search_field.value = None
            self.done_search()

    @property
    def disabled(self):
        return self.dropdown.disabled
    
    @disabled.setter
    def disabled(self, disabled):
        try:
            self.dropdown.disabled = disabled
            self.dropdown.update()
            self.search_button.disabled = disabled
            self.search_button.update()
            self.search_field.disabled = disabled
            self.search_field.update()
            self.search_box.disabled = disabled
            self.done_search()
        except (AttributeError, AssertionError):
            pass
    
    @property
    def options(self):
        return self.dropdown.options
    
    @options.setter
    def options(self, options):
        self._options = options
        self.dropdown.options = self._options

    @property
    def on_change(self):
        return self.dropdown.on_change
    
    @on_change.setter
    def on_change(self, on_change):
        self.dropdown.on_change = on_change

    def update_option_list(self, input: str):
        new_list_options = []
        for option in self._options:
            if input.lower() in option.key.lower():
                new_list_options.append(option)
        return new_list_options

    def on_search(self, e):
        if self.search_field.value.replace(' ', ''):
            self.dropdown.options = self.update_option_list(self.search_field.value)
            if len(self.dropdown.options) == 1:
                self.dropdown.value = self.dropdown.options[0].key
                self.on_change(e, label=self.label, value=self.value)
            else:
                self.dropdown.value = None
        else:
            self.dropdown.options = self._options
        self.dropdown.update()
        self.on_change()

    def search_now(self, e):
        self.search_box.width = self.search_width
        self.search_box.update()
        self.search_button.icon = 'highlight_remove'
        self.search_button.on_click = self.done_search
        self.search_button.update()
        self.search_field.border = "outline"
        self.search_field.update()
        self.search_field.focus()
        if self.search_field.value:
            self.on_search(e)
    
    def done_search(self, e=None):
        self.search_box.width = 0
        self.search_box.update()
        self.search_button.icon = 'search'
        self.search_button.on_click = self.search_now
        self.search_button.update()
        self.search_field.border = "none"
        self.search_field.update()
        self.options = self._options
        self.dropdown.update()


class Collapsible(ft.Column):
    def __init__(
        self,
        title: str,
        content: ft.Control,
        icon: Optional[ft.Control] = None,
        spacing: float = 3,
        radio: Optional[ft.Radio] = None,
        scale: ft.types.ScaleValue = 1,
    ):
        super().__init__()
        self.icon = icon
        self.title = title
        self.scale = scale
        self.shevron = ft.Icon(
            ft.icons.KEYBOARD_ARROW_RIGHT_ROUNDED,
            animate_rotation=100,
            rotate=0,
            scale=self.scale,
        )
        self.content = ft.Column(
            [ft.Container(height=spacing), content],
            height=0,
            spacing=0,
            animate_size=100,
            opacity=0,
            animate_opacity=100,
            scale=self.scale,
        )
        self.spacing = 0
        self.radio = radio
        self.radio.scale = self.scale

    def header_click(self, e):
        self.content.height = None if self.content.height == 0 else 0
        self.content.opacity = 0 if self.content.height == 0 else 1
        self.shevron.rotate = pi / 2 if self.shevron.rotate == 0 else 0
        self.update()

    def build(self):
        title_row = ft.Row()
        if self.icon is not None:
            title_row.controls.append(self.icon)
        if self.radio:
            title_row.controls.append(self.radio)
        else:
            title_row.controls.append(ft.Text(self.title))
        self.controls = [
            ft.Container(
                ft.Row([title_row, self.shevron], alignment="spaceBetween"),
                padding=ft.padding.only(left=8, right=8),
                height=38,
                border_radius=4,
                ink=True,
                on_click=self.header_click,
            ),
            self.content,
        ]

        
class MenuButton(ft.Container):
    def __init__(
        self,
        title: str,
        icon: Optional[ft.Control] = None,
        selected: bool = False,
        radio: Optional[ft.Radio] = None,
    ):
        super().__init__()
        self.icon = icon
        self.title = title
        self._selected = selected
        self.padding = ft.padding.only(left=43)
        self.height = 38
        self.border_radius = 4
        self.ink = True
        self.on_click = self.item_click
        self.radio = radio

    def item_click(self, _):
        pass

    def build(self):
        row = ft.Row()
        if self.icon is not None:
            row.controls.append(self.icon)
        if self.radio:
            row.controls.append(self.radio)
        else:
            row.controls.append(ft.Text(self.title))
        self.content = row

    def _before_build_command(self):
        self.bgcolor = "surfacevariant" if self._selected else None
        super()._before_build_command()



from ..common.tools import download

SEARCH_HEADERS = [
    'productDescEn',
    'productIntroEn',
    'productCode',
    'brandNameEn',
    'productModel',
    'pdfUrl',
    'productImages',
]
PARAMETERS_MAP = [
    'paramVOList',
    'paramNameEn',
    'paramValueEn',
]

PRICING_MAP = [
    'productPriceList',
    'ladder',
    'usdPrice',
]


def get_default_search_keys():
    return [
        'productIntroEn',
        'productIntroEn',
        'revision',
        'keywords',
        'productCode',
        'brandNameEn',
        'productModel',
        '',
        'pdfUrl',
        'productImages',
    ]


def find_categories(part_details: str):
    ''' Find categories '''
    try:
        return part_details['parentCatalogName'], part_details['catalogName']
    except:
        return None, None


def fetch_part_info(part_number: str) -> dict:
    ''' Fetch part data from API '''

    # Load LCSC settings
    from ..config import settings, config_interface
    lcsc_api_settings = config_interface.load_file(settings.CONFIG_LCSC_API)

    part_info = {}

    def search_timeout(timeout=10):
        url = lcsc_api_settings.get('LCSC_API_URL', '') + part_number
        response = download(url, timeout=timeout)
        return response

    # Query part number
    try:
        part = search_timeout()
    except:
        part = None

    # Extract result
    part = part.get('result', None)

    if not part:
        return part_info

    category, subcategory = find_categories(part)
    try:
        part_info['category'] = category
        part_info['subcategory'] = subcategory
    except:
        part_info['category'] = ''
        part_info['subcategory'] = ''

    headers = SEARCH_HEADERS

    for key in part:
        if key in headers:
            if key == 'productImages':
                try:
                    part_info[key] = part['productImages'][0]
                except IndexError:
                    pass
            else:
                part_info[key] = part[key]

    # Parameters
    part_info['parameters'] = {}
    [parameter_key, name_key, value_key] = PARAMETERS_MAP

    if part.get(parameter_key, ''):
        for parameter in range(len(part[parameter_key])):
            parameter_name = part[parameter_key][parameter][name_key]
            parameter_value = part[parameter_key][parameter][value_key]
            # Append to parameters dictionary
            part_info['parameters'][parameter_name] = parameter_value

    # Pricing
    part_info['pricing'] = {}
    [pricing_key, qty_key, price_key] = PRICING_MAP

    for price_break in part[pricing_key]:
        quantity = price_break[qty_key]
        price = price_break[price_key]
        part_info['pricing'][quantity] = price

    part_info['currency'] = 'USD'

    # Extra search fields
    if settings.CONFIG_LCSC.get('EXTRA_FIELDS', None):
        for extra_field in settings.CONFIG_LCSC['EXTRA_FIELDS']:
            if part.get(extra_field, None):
                part_info['parameters'][extra_field] = part[extra_field]
            else:
                from ..common.tools import cprint
                cprint(f'[INFO]\tWarning: Extra field "{extra_field}" not found in search results', silent=False)

    return part_info


def test_api() -> bool:
    ''' Test method for API '''

    test_success = True
    expected = {
        'productIntroEn': '25V 100pF C0G ±5% 0201  Multilayer Ceramic Capacitors MLCC - SMD/SMT ROHS',
        'productCode': 'C2181718',
        'brandNameEn': 'TDK',
        'productModel': 'C0603C0G1E101J030BA',
    }

    test_part = fetch_part_info('C2181718')
    if not test_part:
        test_success = False
        
    # Check content of response
    if test_success:
        for key, value in expected.items():
            if test_part[key] != value:
                print(f'"{test_part[key]}" <> "{value}"')
                test_success = False
                break

    return test_success



from ..config import settings
import validators
from ..common import part_tools
from ..common.tools import cprint, download_with_retry
from ..config import config_interface
import re

# Required to use local CA certificates on Linux
# For more details, refer to https://github.com/sparkmicro/Ki-nTree/pull/45
import platform
import os
if platform.system() == 'Linux':
    cert_path = '/etc/ssl/certs/ca-certificates.crt'
    if os.path.isfile(cert_path):
        os.environ['REQUESTS_CA_BUNDLE'] = cert_path

# InvenTree
from inventree.api import InvenTreeAPI
from inventree.company import Company, ManufacturerPart, SupplierPart, SupplierPriceBreak
from inventree.part import Part, PartCategory, Parameter, ParameterTemplate
from inventree.currency import CurrencyManager
from inventree.stock import StockLocation
from inventree.stock import StockItem


def connect(server: str,
            username: str,
            password: str,
            connect_timeout=5,
            silent=False,
            proxies=None,
            token='') -> bool:
    ''' Connect to InvenTree server and create API object '''
    from wrapt_timeout_decorator import timeout
    global inventree_api

    @timeout(dec_timeout=connect_timeout)
    def get_inventree_api_timeout():
        return InvenTreeAPI(server,
                            username=username,
                            password=password,
                            proxies=proxies,
                            token=token)

    try:
        inventree_api = get_inventree_api_timeout()
    except:
        return False

    if inventree_api.token:
        return True
    return False


def set_inventree_db_test_mode():
    ''' InvenTree test database setup '''
    global inventree_api

    inventree_api.patch('settings/global/PART_PARAMETER_ENFORCE_UNITS', {'value': False})


def get_inventree_category_id(category_tree: list) -> int:
    ''' Get InvenTree category ID from name, specificy parent if subcategory '''
    global inventree_api

    # Fetch all categories
    part_categories = PartCategory.list(inventree_api, name=category_tree[-1])
    if len(part_categories) == 1:
        return part_categories[0].pk
    else:
        if len(category_tree) > 1:
            # Match the parent category
            parent_category_id = get_inventree_category_id(category_tree[:-1])
            if parent_category_id:
                for category in part_categories:
                    try:
                        if parent_category_id == category.getParentCategory().pk:
                            return category.pk
                    except AttributeError:
                        pass
                    #     # Check parent id match (if passed as argument)
                    #     match = True
                    #     if parent_category_id:
                    #         cprint(f'[TREE]\t{item.getParentCategory().pk} ?= {parent_category_id}', silent=settings.HIDE_DEBUG)
                    #         if item.getParentCategory().pk != parent_category_id:
                    #             match = False
                    #     if match:
                    #         cprint(f'[TREE]\t{item.name} ?= {category_name} => True', silent=settings.HIDE_DEBUG)
                    #         return item.pk
                    # else:
                    #     cprint(f'[TREE]\t{item.name} ?= {category_name} => False', silent=settings.HIDE_DEBUG)

    return -1


def get_inventree_stock_location_id(stock_location_tree: list) -> int:
    ''' Get InvenTree stock location ID from name, specificy parent if subcategory '''
    global inventree_api

    # Fetch all categories
    stock_locations = StockLocation.list(inventree_api, name=stock_location_tree[-1])
    if len(stock_locations) == 1:
        return stock_locations[0].pk
    else:
        if len(stock_location_tree) > 1:
            # Match the parent category
            parent_stock_location_id = get_inventree_category_id(stock_location_tree[:-1])
            if parent_stock_location_id:
                for location in stock_locations:
                    try:
                        if parent_stock_location_id == location.getParentLocation().pk:
                            return location.pk
                    except AttributeError:
                        pass
                    #     # Check parent id match (if passed as argument)
                    #     match = True
                    #     if parent_stock_location_id:
                    #         cprint(f'[TREE]\t{item.getParentCategory().pk} ?= {parent_stock_location_id}', silent=settings.HIDE_DEBUG)
                    #         if item.getParentCategory().pk != parent_stock_location_id:
                    #             match = False
                    #     if match:
                    #         cprint(f'[TREE]\t{item.name} ?= {category_name} => True', silent=settings.HIDE_DEBUG)
                    #         return item.pk
                    # else:
                    #     cprint(f'[TREE]\t{item.name} ?= {category_name} => False', silent=settings.HIDE_DEBUG)

    return -1


def get_categories() -> dict:
    '''Fetch InvenTree categories'''
    global inventree_api

    categories = {}
    # Get all categories (list)
    db_categories = PartCategory.list(inventree_api)

    def deep_add(tree: dict, keys: list, item: dict):
        if len(keys) == 1:
            try:
                tree[keys[0]].update(item)
            except (KeyError, AttributeError):
                tree[keys[0]] = item
            return
        return deep_add(tree.get(keys[0]), keys[1:], item)

    for category in db_categories:
        parent = category.getParentCategory()
        children = category.getChildCategories()

        if not parent and not children:
            categories[category.name] = None
            continue
        elif parent:
            parent_list = []
            while parent:
                parent_list.insert(0, parent.name)
                parent = parent.getParentCategory()
            cat = {category.name: None}
            deep_add(categories, parent_list, cat)

    return categories


def get_stock_locations() -> dict:
    '''Fetch InvenTree stock locations'''
    global inventree_api

    categories = {}
    # Get all categories (list)
    db_categories = StockLocation.list(inventree_api)

    def deep_add(tree: dict, keys: list, item: dict):
        if len(keys) == 1:
            try:
                tree[keys[0]].update(item)
            except (KeyError, AttributeError):
                tree[keys[0]] = item
            return
        return deep_add(tree.get(keys[0]), keys[1:], item)

    for category in db_categories:
        parent = category.getParentLocation()
        children = category.getChildLocations()

        if not parent and not children:
            categories[category.name] = None
            continue
        elif parent:
            parent_list = []
            while parent:
                parent_list.insert(0, parent.name)
                parent = parent.getParentLocation()
            cat = {category.name: None}
            deep_add(categories, parent_list, cat)

    return categories


def get_category_tree(category_id: int) -> dict:
    ''' Get all parents of a category'''
    category = PartCategory(inventree_api, category_id)
    category_list = {category_id: category.name}

    while category.parent:
        category = category.getParentCategory()
        category_list[category.pk] = category.name

    return category_list


def get_stock_location_tree(id: int) -> dict:
    ''' Get all parents of a stock_location'''
    location = StockLocation(inventree_api, id)
    list = {id: location.name}

    while location.parent:
        location = location.getParentLocation()
        list[location.pk] = location.name

    return list


def create_stock(stock_data: dict) -> dict:
    return StockItem.create(inventree_api, stock_data)


def get_category_parameters(category_id: int) -> list:
    ''' Get all default parameter templates for category '''
    global inventree_api

    parameter_templates = []

    category = PartCategory(inventree_api, category_id)

    try:
        category_templates = category.getCategoryParameterTemplates(fetch_parent=True)
    except AttributeError:
        category_templates = None

    if category_templates:
        for template in category_templates:

            default_value = template.default_value
            if not default_value:
                default_value = '-'

            parameter_templates.append([template.getTemplate().name, default_value])

    return parameter_templates


def get_part_info(part_id: int) -> str:
    ''' Get InvenTree part info from specified Part ID '''
    global inventree_api

    part = Part(inventree_api, part_id)
    part_info = {'IPN': part.IPN}
    attachment = part.getAttachments()
    if attachment:
        part_info['datasheet'] = f'{inventree_api.base_url.strip("/")}{attachment[0]["attachment"]}'
    return part_info


def set_part_number(part_id: int, ipn: str) -> bool:
    ''' Set InvenTree part number for specified Part ID '''
    data = {'IPN': ipn}
    update_part(part_id, data)

    if Part(inventree_api, part_id).IPN == ipn:
        return True
    else:
        return False


def get_part_from_ipn(part_ipn='') -> int:
    ''' Get Part ID from Part IPN '''
    global inventree_api

    parts = Part.list(inventree_api, IPN=part_ipn)

    if not parts:
        # No part found
        return None
    else:
        # parts should have only one entry
        return parts[0]


def fetch_part(part_id='', part_ipn='') -> int:
    ''' Fetch part from database using either ID or IPN '''
    from requests.exceptions import HTTPError
    global inventree_api

    part = None
    if part_id:
        try:
            part = Part(inventree_api, part_id)
        except TypeError:
            # Part ID is invalid (eg. decimal value)
            cprint('[TREE] Error: Part ID type is invalid')
        except ValueError:
            # Part ID is not a positive integer
            cprint('[TREE] Error: Part ID must be positive')
        except HTTPError:
            # Part ID does not exist
            cprint(f'[TREE] Error: Part with ID={part_id} does not exist in database')
    elif part_ipn:
        part = get_part_from_ipn(part_ipn)
    else:
        pass

    return part


def is_new_part(category_id: int, part_info: dict) -> int:
    ''' Check if part exists based on parameters (or description) '''
    global inventree_api

    # Get category object
    part_category = PartCategory(inventree_api, category_id)

    # Fetch all parts from category and subcategories
    part_list = []
    part_list.extend(part_category.getParts())
    for subcategory in part_category.getChildCategories():
        part_list.extend(subcategory.getParts())

    # Extract parameter from part info
    # Verify parameters values are not empty
    new_part_parameters = part_info['parameters'] if list(set(part_info['parameters'].values())) != ['-'] else None

    template_list = ParameterTemplate.list(inventree_api)

    def fetch_template_name(template_id):
        for item in template_list:
            if item.pk == template_id:
                return item.name

    # Retrieve parent category name for parameters compare
    try:
        category_name = part_category.getParentCategory().name
    except AttributeError:
        category_name = part_category.name
    filters = config_interface.load_category_parameters_filters(category=category_name,
                                                                supplier_config_path=settings.CONFIG_PARAMETERS_FILTERS)
    # cprint(filters)

    for part in part_list:
        # TODO: This statement below seems erroneous...
        # Compare fields (InvenTree does not allow those to be identicals between two parts)
        # compare_fields = part_info['name'] == part.name and part_info['revision'] == part.revision
        # if compare_fields:
        #     cprint(f'[TREE]\tWarning: Found part with same name and revision (pk = {part.pk})', silent=settings.SILENT)
        #     return part.pk

        # Compare parameters
        compare_parameters = False
        # Get part parameters
        db_part_parameters = part.getParameters()
        part_parameters = {}
        for parameter in db_part_parameters:
            parameter_name = fetch_template_name(parameter.template)
            parameter_value = parameter.data
            part_parameters[parameter_name] = parameter_value

        if new_part_parameters and part_parameters:
            # Compare database part with new part
            compare_parameters = part_tools.compare(new_part_parameters=new_part_parameters,
                                                    db_part_parameters=part_parameters,
                                                    include_filters=filters)
                                                            
        if compare_parameters:
            cprint(f'[TREE]\tWarning: Found part with same parameters in database (pk = {part.pk})', silent=settings.SILENT)
            return part.pk

    # Check if manufacturer part exists in database
    manufacturer = part_info['manufacturer_name']
    mpn = part_info['manufacturer_part_number']
    part_pk = is_new_manufacturer_part(manufacturer, mpn, create=False)

    if part_pk:
        cprint(f'[TREE]\tWarning: Found part with same manufacturer and MPN in database (pk = {part_pk})', silent=settings.SILENT)
        return part_pk

    cprint('\n[TREE]\tNo match found in database', silent=settings.HIDE_DEBUG)
    return 0


def create_category(parent: str, name: str):
    ''' Create InvenTree category, use parent for subcategories '''
    global inventree_api

    parent_id = 0
    is_new_category = False

    # Check if category already exists
    category_list = PartCategory.list(inventree_api)
    for category in category_list:
        if name == category.name:
            try:
                # Check if parents are the same
                if category.getParentCategory().name == parent:
                    # Return category ID
                    return category.pk, is_new_category
            except:
                return category.pk, is_new_category
        elif parent == category.name:
            # Get Parent ID
            parent_id = category.pk
        else:
            pass

    if parent:
        if parent_id > 0:
            category = PartCategory.create(inventree_api, {
                'name': name,
                'parent': parent_id,
            })

            is_new_category = True
        else:
            cprint(f'[TREE]\tError: Check parent category name ({parent})', silent=settings.SILENT)
            return -1, is_new_category
    else:
        # No parent
        category = PartCategory.create(inventree_api, {
            'name': name,
            'parent': None,
        })
        is_new_category = True

    try:
        category_pk = category.pk
    except AttributeError:
        # User does not have the permission to create categories
        category_pk = 0

    return category_pk, is_new_category


def upload_part_image(image_url: str, part_id: int) -> bool:
    ''' Upload InvenTree part thumbnail'''
    global inventree_api

    # Get image full path
    image_name = f'{str(part_id)}_thumbnail.jpeg'
    image_location = settings.search_images + image_name

    # Download image (multiple attempts)
    if not download_with_retry(image_url, image_location, filetype='Image'):
        return False

    # Upload image to InvenTree
    part = Part(inventree_api, part_id)
    if part:
        try:
            return part.uploadImage(image=image_location)
        except Exception:
            return False
    else:
        return False


def upload_part_datasheet(datasheet_url: str, part_ipn: int, part_pk: int) -> str:
    ''' Upload InvenTree part attachment'''
    global inventree_api

    datasheet_name = f'{part_ipn}.pdf'
    # Get datasheet path based on user settings for local storage
    if settings.DATASHEET_SAVE_ENABLED:
        datasheet_location = os.path.join(settings.DATASHEET_SAVE_PATH, datasheet_name)
    else:
        datasheet_location = os.path.join(settings.search_datasheets, datasheet_name)

    if not os.path.isfile(datasheet_location):
        # Download datasheet (multiple attempts)
        if not download_with_retry(
            datasheet_url,
            datasheet_location,
            filetype='PDF',
            timeout=10
        ):
            return ''

    # Upload Datasheet to InvenTree
    part = Part(inventree_api, part_pk)
    if part:
        try:
            attachment = part.uploadAttachment(attachment=datasheet_location)
            return f'{inventree_api.base_url.strip("/")}{attachment["attachment"]}'
        except Exception:
            return ''
    else:
        return ''


def create_part(category_id: int, name: str, description: str, revision: str, ipn: str, keywords=None) -> int:
    ''' Create InvenTree part '''
    global inventree_api

    try:
        part = Part.create(inventree_api, {
            'name': name,
            'description': description,
            'category': category_id,
            'keywords': keywords,
            'revision': revision,
            'IPN': ipn,
            'active': True,
            'virtual': False,
            'component': True,
            'purchaseable': True,
        })
    except Exception:
        cprint('[TREE]\tError: Part creation failed. Check if Ki-nTree settings match InvenTree part settings.', silent=settings.SILENT)
        return 0

    if part:
        return part.pk
    else:
        return 0


def set_part_default_location(part_pk: int, location_pk: int):
    global inventree_api

    # Retrieve part instance with primary-key of 1
    part = Part(inventree_api, pk=part_pk)

    # Update specified part parameters
    part.save(data={
        "default_location": location_pk,
    })


def update_part(pk: int, data: dict) -> int:
    '''Update an existing parts data'''
    global inventree_api

    part = Part(inventree_api, pk)
    if part:
        part.save(data=data)
        return part.pk
    else:
        return 0


def create_company(company_name: str, manufacturer=False, supplier=False) -> bool:
    ''' Create InvenTree company '''
    global inventree_api

    if not manufacturer and not supplier:
        return None

    company = Company.create(inventree_api, {
        'name': company_name,
        'description': company_name,
        'is_customer': False,
        'is_supplier': supplier,
        'is_manufacturer': manufacturer,
    })

    return company


def get_all_companies() -> dict:
    ''' Get all existing companies (supplier/manufacturer) from database '''
    global inventree_api

    company_list = Company.list(inventree_api)
    companies = {}
    for company in company_list:
        companies[company.name] = company.pk

    return companies


def get_company_id(company_name: str) -> int:
    ''' Get company (supplier/manufacturer) primary key (ID) '''

    try:
        return get_all_companies()[company_name]
    except:
        return 0


def is_new_manufacturer_part(manufacturer_name: str, manufacturer_mpn: str, create=True) -> int:
    ''' Check if InvenTree manufacturer part exists to avoid duplicates '''
    global inventree_api

    if not manufacturer_name:
        return 0

    # Fetch all companies
    cprint('[TREE]\tFetching manufacturers', silent=settings.HIDE_DEBUG)
    company_list = Company.list(inventree_api, is_manufacturer=True, is_customer=False)
    companies = {}
    for company in company_list:
        companies[company.name] = company

    try:
        # Get all parts
        part_list = companies[manufacturer_name].getManufacturedParts()
    except:
        part_list = None

    if part_list is None:
        if create:
            # Create manufacturer
            cprint(f'[TREE]\tCreating new manufacturer "{manufacturer_name}"', silent=settings.SILENT)
            create_company(
                company_name=manufacturer_name,
                manufacturer=True,
            )
        # Get all parts
        part_list = []

    for item in part_list:
        try:
            if manufacturer_mpn in item.MPN:
                cprint(f'[TREE]\t{item.MPN} ?= {manufacturer_mpn} => True', silent=settings.HIDE_DEBUG)
                return item.part
            else:
                cprint(f'[TREE]\t{item.MPN} ?= {manufacturer_mpn} => False', silent=settings.HIDE_DEBUG)
        except TypeError:
            cprint(f'[TREE]\t{item.MPN} ?= {manufacturer_mpn} => *** SKIPPED ***', silent=settings.HIDE_DEBUG)

    return 0


def is_new_supplier_part(supplier_name: str, supplier_sku: str):
    ''' Check if InvenTree supplier part exists to avoid duplicates '''
    global inventree_api

    # Fetch all companies
    cprint('[TREE]\tFetching suppliers', silent=settings.HIDE_DEBUG)
    company_list = Company.list(inventree_api, is_supplier=True, is_customer=False)
    companies = {}
    for company in company_list:
        companies[company.name] = company

    try:
        # Get all parts
        part_list = companies[supplier_name].getSuppliedParts()
    except:
        part_list = None

    if part_list is None:
        # Create
        cprint(f'[TREE]\tCreating new supplier "{supplier_name}"', silent=settings.SILENT)
        create_company(
            company_name=supplier_name,
            supplier=True,
        )
        # Get all parts
        part_list = []

    for item in part_list:
        if supplier_sku in item.SKU:
            cprint(f'[TREE]\t{item.SKU} ?= {supplier_sku} => True', silent=settings.HIDE_DEBUG)
            return False, item
        else:
            cprint(f'[TREE]\t{item.SKU} ?= {supplier_sku} => False', silent=settings.HIDE_DEBUG)

    return True, False


def create_manufacturer_part(part_id: int, manufacturer_name: str, manufacturer_mpn: str, description: str, datasheet: str) -> bool:
    ''' Create InvenTree manufacturer part

        part_id: Part the manufacturer data is linked to
        manufacturer: Company that manufactures the SupplierPart (leave blank if it is the sample as the Supplier!)
        MPN: Manufacture part number
        datasheet: Datasheet link
        description: Descriptive notes field
    '''
    global inventree_api

    # Get Manufacturer ID
    manufacturer_id = get_company_id(manufacturer_name)

    if manufacturer_id:
        # Validate datasheet link
        if not validators.url(datasheet):
            datasheet = ''

        manufacturer_part = ManufacturerPart.create(inventree_api, {
            'part': part_id,
            'manufacturer': manufacturer_id,
            'MPN': manufacturer_mpn,
            'link': datasheet,
            'description': description,
        })

        if manufacturer_part:
            return True
    else:
        cprint(f'[TREE]\tError: Manufacturer "{manufacturer_name}" not found (failed to create manufacturer part)',
               silent=settings.SILENT)

    return False


def create_supplier_part(part_id: int, manufacturer_name: str, manufacturer_mpn: str, supplier_name: str, supplier_sku: str, description: str, link: str):
    ''' Create InvenTree supplier part

        part_id: Part the supplier data is linked to
        manufacturer_name: Manufacturer the supplier data is linked to
        manufacturer_mpn: MPN the supplier data is linked to
        supplier: Company that supplies this SupplierPart object
        SKU: Stock keeping unit (supplier part number)
        manufacturer: Company that manufactures the SupplierPart (leave blank if it is the sample as the Supplier!)
        MPN: Manufacture part number
        link: Link to part detail page on supplier's website
        description: Descriptive notes field
    '''
    global inventree_api

    # Get Supplier ID
    supplier_id = get_company_id(supplier_name)

    if not manufacturer_name or not manufacturer_mpn:
        # Unset manufacturer data
        manufacturer_name = None
        manufacturer_mpn = None

    if supplier_id:
        # Validate supplier link
        if not validators.url(link):
            link = ''

        supplier_part = SupplierPart.create(inventree_api, {
            'part': part_id,
            'manufacturer': manufacturer_name,
            'MPN': manufacturer_mpn,
            'supplier': supplier_id,
            'SKU': supplier_sku,
            'link': link,
            'description': description,
        })

        if supplier_part:
            return True, supplier_part
    else:
        cprint(f'[TREE]\tError: Supplier "{supplier_name}" not found (failed to create supplier part)',
               silent=settings.SILENT)

    return False, False


def sanitize_price(price_in):
    price = re.findall('\d+.\d+', price_in)[0]
    price = price.replace(',', '.')
    price = price.replace('\xa0', '')
    return price


def update_price_breaks(supplier_part,
                        price_breaks: dict,
                        currency='USD') -> bool:
    ''' Update the Price Breaks associated with a supplier part '''
    def sanitize_price(price_in):
        price = re.findall('\d+.\d+', price_in)[0]
        price = price.replace(',', '.')
        price = price.replace('\xa0', '')
        return price

    def convert_currency(price):
        manager = CurrencyManager(inventree_api)
        base = manager.getBaseCurrency()
        if base != currency:
            try:
                price = manager.convertCurrency(float(price), currency, base)
            except Exception:
                cprint('[TREE]\tWarning: Currency conversion failed.',
                       silent=settings.SILENT)
        return price

    if not isinstance(supplier_part, SupplierPart):
        try:
            supplier_part = SupplierPart(inventree_api, supplier_part)
        except:
            cprint('[TREE]\tWarning: Supplier part not found, skipping price break update',
                   silent=settings.SILENT)
            return False
    if not price_breaks:
        cprint('[TREE]\tWarning: No price breaks found, skipping.', silent=settings.SILENT)
        return False

    old_price_breaks = supplier_part.getPriceBreaks()
    updated = []
    # First process existing price breaks
    for old_price_break in old_price_breaks:
        quantity = old_price_break.quantity
        if quantity in price_breaks:
            price = price_breaks[quantity]
            # remove everything but the numbers from the price break
            if isinstance(price, str):
                price = sanitize_price(price)
            price = convert_currency(price)
            old_price_break.save(data={'price': price})
            updated.append(quantity)
        else:
            old_price_break.delete()
    for quantity in updated:
        del price_breaks[quantity]
    # if any price breaks are left over these will be created
    for quantity, price in price_breaks.items():
        # remove everything but the numbers from the price break
        if isinstance(price, str):
            price = sanitize_price(price)
        price = convert_currency(price)
        SupplierPriceBreak.create(inventree_api, {
            'part': supplier_part.pk,
            'quantity': quantity,
            'price': price,
        })
    cprint('[INFO]\tSuccess: The price breaks were updated', silent=settings.SILENT)
    return True


def create_parameter_template(name: str, units: str) -> int:
    ''' Create InvenTree parameter template '''
    global inventree_api

    parameter_templates = ParameterTemplate.list(inventree_api)
    for item in parameter_templates:
        if name == item.name:
            return 0

    try:
        parameter_template = ParameterTemplate.create(inventree_api, {
            'name': name,
            'units': units if units else '',
        })
    except:
        cprint(f'[TREE]\tError: Failed to create parameter template "{name}".', silent=settings.SILENT)

    if parameter_template:
        return parameter_template.pk
    else:
        return 0


def create_parameter(part_id: int, template_name: int, value: str):
    ''' Create InvenTree part parameter based on template '''
    global inventree_api

    parameter_template_list = ParameterTemplate.list(inventree_api)

    template_id = 0
    for item in parameter_template_list:
        if template_name == item.name:
            template_id = item.pk
            break

    # Check if template_id already exists for this part
    part = Part(inventree_api, part_id)
    part_parameters = part.getParameters()
    is_new_part_parameters_template_id = True
    was_updated = False
    parameter = None
    for item in part_parameters:
        # cprint(f'[TREE]\t{parameter.template} ?= {template_id}', silent=SILENT)
        if item.template == template_id:
            is_new_part_parameters_template_id = False
            if settings.UPDATE_INVENTREE:
                if value != item.data and value != '-':
                    parameter = item
                    was_updated = True
                    try:
                        parameter.save(data={
                            'data': value
                        })
                    except Exception as e:
                        cprint(f'[TREE]\tError: Failed to update part parameter "{template_name}".', silent=settings.SILENT)
                        if "Could not convert" in e.args[0]['body'].__str__():
                            cprint(f'[TREE]\tError: Parameter value "{value}" is not allowed by server settings.', silent=settings.SILENT)
            break
    # cprint(part_parameters, silent=SILENT)

    '''
        Create parameter only if:
        - template exists
        - parameter does not exist for this part
    '''
    if template_id > 0 and is_new_part_parameters_template_id:
        try:
            parameter = Parameter.create(inventree_api, {
                'part': part_id,
                'template': template_id,
                'data': value,
            })
        except Exception as e:
            cprint(f'[TREE]\tError: Failed to create part parameter "{template_name}".', silent=settings.SILENT)
            if "Could not convert" in e.args[0]['body'].__str__():
                cprint(f'[TREE]\tError: Parameter value "{value}" is not allowed by server settings.', silent=settings.SILENT)

    if parameter:
        return parameter.pk, is_new_part_parameters_template_id, was_updated
    else:
        if template_id == 0:
            cprint(f'[TREE]\tError: Parameter template "{template_name}" does not exist', silent=settings.SILENT)
        return 0, False, False



import flet as ft

# Common view
from .common import DialogType
from .common import CommonView
from .common import SwitchWithRefs
from .common import GUI_PARAMS
from .common import handle_transition
# Settings
from ...config import settings as global_settings
from ...config import config_interface


# Load Supplier Settings
supplier_settings = {}
for supplier, data in global_settings.CONFIG_SUPPLIERS.items():
    supplier_settings[supplier] = {}

    # Add enable
    supplier_settings[supplier]['Enable'] = [
        data['enable'],
        ft.Switch(),
        None,
    ]

    # Add supplier name
    supplier_settings[supplier]['InvenTree Name'] = [
        data['name'],
        ft.TextField(),
        None,
    ]

    # Add API fields
    if supplier == 'Digi-Key':
        digikey_api_settings = config_interface.load_file(global_settings.CONFIG_DIGIKEY_API)
        supplier_settings[supplier]['Client ID'] = [
            digikey_api_settings['DIGIKEY_CLIENT_ID'],
            ft.TextField(),
            None,
        ]
        supplier_settings[supplier]['Client Secret'] = [
            digikey_api_settings['DIGIKEY_CLIENT_SECRET'],
            ft.TextField(),
            None,
        ]
    elif supplier == 'Mouser':
        mouser_api_settings = config_interface.load_file(global_settings.CONFIG_MOUSER_API)
        supplier_settings[supplier]['Part API Key'] = [
            mouser_api_settings['MOUSER_PART_API_KEY'],
            ft.TextField(),
            None,
        ]
    elif supplier == 'Element14' or supplier == 'Farnell' or supplier == 'Newark':
        from ...search.element14_api import STORES
        element14_api_settings = config_interface.load_file(global_settings.CONFIG_ELEMENT14_API)
        default_store = element14_api_settings.get(f'{supplier.upper()}_STORE', '')

        supplier_settings[supplier]['Product Search API Key (Element14)'] = [
            element14_api_settings['ELEMENT14_PRODUCT_SEARCH_API_KEY'],
            ft.TextField(),
            None,
        ]
        
        dropdown_options = []
        for store_name, store_url in STORES[supplier].items():
            dropdown_options.append(ft.dropdown.Option(f'{store_name} ({store_url})'))
        supplier_settings[supplier][f'{supplier} Store'] = [
            default_store,
            ft.Dropdown(
                label='Store',
                width=GUI_PARAMS['dropdown_width'],
                dense=GUI_PARAMS['dropdown_dense'],
                options=dropdown_options
            ),
            None,
        ]
    elif supplier == 'LCSC':
        lcsc_api_settings = config_interface.load_file(global_settings.CONFIG_LCSC_API)
        supplier_settings[supplier]['API URL'] = [
            lcsc_api_settings['LCSC_API_URL'],
            ft.TextField(),
            None,
        ]
    elif supplier == 'TME':
        tme_api_settings = config_interface.load_file(global_settings.CONFIG_TME_API)
        supplier_settings[supplier]['API Token'] = [
            tme_api_settings['TME_API_TOKEN'],
            ft.TextField(),
            None,
        ]
        supplier_settings[supplier]['API Secret'] = [
            tme_api_settings['TME_API_SECRET'],
            ft.TextField(),
            None,
        ]
        supplier_settings[supplier]['API Country'] = [
            tme_api_settings['TME_API_COUNTRY'],
            ft.TextField(),
            None,
        ]
        supplier_settings[supplier]['API Language'] = [
            tme_api_settings['TME_API_LANGUAGE'],
            ft.TextField(),
            None,
        ]

SETTINGS = {
    'User Settings': {
        'Configuration Files Folder': [
            'USER_FILES',
            ft.TextField(),
            True,  # Browse enabled
        ],
        'Cache Folder': [
            'USER_CACHE',
            ft.TextField(),
            True,  # Browse enabled
        ],
        'Save Datasheets to Local Folder': [
            'DATASHEET_SAVE_ENABLED',
            SwitchWithRefs(),
            False,  # Browse enabled
        ],
        'Datasheet Folder': [
            'DATASHEET_SAVE_PATH',
            ft.TextField(),
            True,  # Browse enabled
        ],
        'Open Browser After Creating Part': [
            'AUTOMATIC_BROWSER_OPEN',
            ft.Switch(),
            False,  # Browse enabled
        ],
        'Enable Supplier Search Cache': [
            'CACHE_ENABLED',
            SwitchWithRefs(),
            False,  # Browse enabled
        ],
        'CACHE_VALID_DAYS': [
            'CACHE_VALID_DAYS',
            ft.TextField(
                text_align=ft.TextAlign.CENTER,
                width=60,
                dense=True,
                disabled=True,
            ),
            False,
        ]
    },
    'Supplier Settings': supplier_settings,
    'InvenTree Settings': {
        'Server Address': [
            'SERVER_ADDRESS',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'Username': [
            'USERNAME',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'Password or Token': [
            'PASSWORD',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'Enable Proxy Support': [
            'ENABLE_PROXY',
            SwitchWithRefs(),
            False,  # Browse disabled
        ],
        'Proxy': [
            'PROXY',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'Upload Datasheets to InvenTree': [
            'DATASHEET_UPLOAD',
            SwitchWithRefs(),
            False,  # Browse enabled
        ],
        'Upload Pricing Data to InvenTree': [
            'PRICING_UPLOAD',
            SwitchWithRefs(),
            False,  # Browse enabled
        ],
        'Default Part Revision': [
            'INVENTREE_DEFAULT_REV',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'Enable Internal Part Number (IPN)': [
            'IPN_ENABLE_CREATE',
            SwitchWithRefs(),
            False,  # Browse disabled
        ],
        'Use Manufacturer Part Number as IPN': [
            'IPN_USE_MANUFACTURER_PART_NUMBER',
            SwitchWithRefs(reverse_dir=True),
            False,  # Browse disabled
        ],
        'IPN: Enable Prefix': [
            'IPN_ENABLE_PREFIX',
            SwitchWithRefs(),
            False,  # Browse disabled
        ],
        'IPN: Prefix': [
            'IPN_PREFIX',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'IPN: Enable Category Codes': [
            'IPN_CATEGORY_CODE',
            ft.Switch(),
            False,  # Browse disabled
        ],
        'IPN: Length of Unique ID': [
            'IPN_UNIQUE_ID_LENGTH',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'IPN: Enable Suffix': [
            'IPN_ENABLE_SUFFIX',
            SwitchWithRefs(),
            False,  # Browse disabled
        ],
        'IPN: Suffix': [
            'IPN_SUFFIX',
            ft.TextField(),
            False,  # Browse disabled
        ],
        'Test': [
            None,
            ft.ElevatedButton,
            False,  # Browse disabled
        ],
    },
    'KiCad Settings': {
        'Symbol Libraries Folder': [
            'KICAD_SYMBOLS_PATH',
            ft.TextField(),
            True,  # Browse enabled
        ],
        'Symbol Templates Folder': [
            'KICAD_TEMPLATES_PATH',
            ft.TextField(),
            True,  # Browse enabled
        ],
        'Footprint Libraries Folder': [
            'KICAD_FOOTPRINTS_PATH',
            ft.TextField(),
            True,  # Browse enabled
        ],
    },
}

# Settings AppBar
settings_appbar = ft.AppBar(
    title=ft.Text('Ki-nTree Settings'),
    bgcolor=ft.colors.SURFACE_VARIANT
)

# Settings NavRail
settings_navrail = ft.NavigationRail(
    selected_index=0,
    label_type=ft.NavigationRailLabelType.ALL,
    min_width=GUI_PARAMS['nav_rail_min_width'],
    min_extended_width=GUI_PARAMS['nav_rail_width'],
    group_alignment=GUI_PARAMS['nav_rail_alignment'],
    destinations=[
        ft.NavigationRailDestination(
            label_content=ft.Text("User", size=GUI_PARAMS['nav_rail_text_size']),
            icon_content=ft.Icon(name=ft.icons.SUPERVISED_USER_CIRCLE, size=GUI_PARAMS['nav_rail_icon_size']),
            selected_icon_content=ft.Icon(name=ft.icons.SUPERVISED_USER_CIRCLE_OUTLINED, size=GUI_PARAMS['nav_rail_icon_size']),
            padding=GUI_PARAMS['nav_rail_padding'],
        ),
        ft.NavigationRailDestination(
            label_content=ft.Text("Supplier", size=GUI_PARAMS['nav_rail_text_size']),
            icon_content=ft.Icon(name=ft.icons.LOCAL_SHIPPING, size=GUI_PARAMS['nav_rail_icon_size']),
            selected_icon_content=ft.Icon(name=ft.icons.LOCAL_SHIPPING_OUTLINED, size=GUI_PARAMS['nav_rail_icon_size']),
            padding=GUI_PARAMS['nav_rail_padding'],
        ),
        ft.NavigationRailDestination(
            label_content=ft.Text("InvenTree", size=GUI_PARAMS['nav_rail_text_size']),
            icon_content=ft.Icon(name=ft.icons.INVENTORY_2, size=GUI_PARAMS['nav_rail_icon_size']),
            selected_icon_content=ft.Icon(name=ft.icons.INVENTORY_2_OUTLINED, size=GUI_PARAMS['nav_rail_icon_size']),
            padding=GUI_PARAMS['nav_rail_padding'],
        ),
        ft.NavigationRailDestination(
            label_content=ft.Text("KiCad", size=GUI_PARAMS['nav_rail_text_size']),
            icon_content=ft.Icon(name=ft.icons.SETTINGS_INPUT_COMPONENT, size=GUI_PARAMS['nav_rail_icon_size']),
            selected_icon_content=ft.Icon(name=ft.icons.SETTINGS_INPUT_COMPONENT_OUTLINED, size=GUI_PARAMS['nav_rail_icon_size']),
            padding=GUI_PARAMS['nav_rail_padding'],
        ),
    ],
    on_change=None,
)

# Navigation indexes (settings)
NAV_BAR_INDEX = {
    0: '/settings/user',
    1: '/settings/supplier',
    2: '/settings/inventree',
    3: '/settings/kicad',
}


class SettingsView(CommonView):
    '''Main settings view'''

    title = 'Settings'
    route = '/settings'
    settings = None
    settings_file = None
    dialog = None

    def __init__(self, page: ft.Page):
        # Load setting fields
        self.fields = {}
        for field_name, field_data in SETTINGS.get(self.title, {}).items():
            if isinstance(field_data, list) and field_data[0] is not None:
                self.fields[field_name] = field_data[1]
                self.fields[field_name].value = self.settings[field_data[0]]

        # Init view
        super().__init__(page=page, appbar=settings_appbar, navigation_rail=settings_navrail)

        # Update navigation rail
        self.navigation_rail.on_change = self.nav_rail_redirect

    def nav_rail_redirect(self, e):
        self._page.go(NAV_BAR_INDEX[e.control.selected_index])
    
    def save(self, settings_file=None, show_dialog=True):
        '''Save settings'''
        if settings_file is not None:
            settings_from_file = config_interface.load_file(settings_file)
        else:
            settings_from_file = config_interface.load_file(self.settings_file)

        # Update settings values
        for key in settings_from_file:
            for setting in SETTINGS[self.title].values():
                if key == setting[0]:
                    settings_from_file[key] = setting[1].value

        # Save
        if settings_file is not None:
            config_interface.dump_file(settings_from_file, settings_file)
        else:
            config_interface.dump_file(settings_from_file, self.settings_file)

        # Alert user
        if show_dialog:
            self.show_dialog(
                d_type=DialogType.VALID,
                message=f'{self.title} successfully saved',
            )

    def on_dialog_result(self, e: ft.FilePickerResultEvent):
        '''Populate field with user-selected system path'''
        if e.path:
            self.fields[e.control.dialog_title].value = e.path
            self._page.update()

    def path_picker(self, e: ft.ControlEvent, title: str):
        '''Let user browse to a system path'''
        if self._page.overlay:
            self._page.overlay.pop()
        path_picker = ft.FilePicker(on_result=self.on_dialog_result)
        self._page.overlay.append(path_picker)
        self._page.update()
        if self.fields[title].value:
            path_picker.get_directory_path(dialog_title=title, initial_directory=self.fields[title].value)
        else:
            path_picker.get_directory_path(dialog_title=title, initial_directory=global_settings.HOME_DIR)

    def init_column(self) -> ft.Column:
        return ft.Column(
            controls=[
                ft.Text(self.title, style="bodyMedium"),
                ft.Row(),
            ],
            alignment=ft.MainAxisAlignment.START,
            expand=True,
        )

    def update_field(self, name, field, column):
        if isinstance(field, ft.TextField):
            field_predefined = bool(field.width)
            if not field_predefined:
                field.label = name
                field.width = GUI_PARAMS['textfield_width']
                field.dense = GUI_PARAMS['textfield_dense']
                if 'password' in field.label.lower():
                    field.password = True
                field_row = ft.Row(
                    controls=[
                        field,
                    ]
                )
                # Add browse button
                if SETTINGS[self.title][name][2]:
                    field_row.controls.append(
                        ft.ElevatedButton(
                            'Browse',
                            width=GUI_PARAMS['button_width'],
                            height=48,
                            on_click=lambda e, t=name: self.path_picker(e, title=t)
                        ),
                    )
                column.controls.extend(
                    [
                        field_row,
                        ft.Row(height=GUI_PARAMS['textfield_space_after']),
                    ]
                )
        elif isinstance(field, ft.Text):
            field.value = name
            field_row = ft.Row(
                controls=[
                    field,
                ]
            )
            column.controls.append(field_row)
            column.controls.append(ft.Divider())
        elif isinstance(field, ft.TextButton):
            column.controls.append(
                ft.ElevatedButton(
                    name,
                    width=GUI_PARAMS['button_width'] * 2,
                    height=GUI_PARAMS['button_height'],
                    icon=ft.icons.CHECK_OUTLINED,
                    on_click=lambda e, s=name: self.test_s(e, s=s)
                ),
            )
        elif isinstance(field, ft.Dropdown):
            field.on_change = lambda _: self.save()
            column.controls.append(
                field,
            )
        elif isinstance(field, ft.Switch) or isinstance(field, SwitchWithRefs):
            if 'proxy' in name.lower():
                field.on_change = lambda _: None
            else:
                field.on_change = lambda _: self.save()
            field.label = name
            column.controls.append(
                field,
            )

    def add_buttons(self, column, test=False) -> ft.Row:
        test_save_buttons = ft.Row()
        if test:
            test_save_buttons.controls.append(
                ft.ElevatedButton(
                    'Test',
                    width=GUI_PARAMS['button_width'],
                    height=GUI_PARAMS['button_height'],
                    icon=ft.icons.CHECK_OUTLINED,
                    on_click=lambda _: self.test(),
                ),
            )
        test_save_buttons.controls.append(
            ft.ElevatedButton(
                'Save',
                width=GUI_PARAMS['button_width'],
                height=GUI_PARAMS['button_height'],
                icon=ft.icons.SAVE_OUTLINED,
                on_click=lambda _: self.save()
            ),
        )
        column.controls.append(test_save_buttons)

    def build_column(self, ignore=[]):
        # Header
        self.column = self.init_column()

        # Fields
        for name, field in self.fields.items():
            if name not in ignore:
                self.update_field(name, field, self.column)

        # Test and Save buttons
        enable_test = bool(list(SETTINGS[self.title])[-1] == 'Test')
        self.add_buttons(self.column, test=enable_test)

    def did_mount(self):
        handle_transition(self._page, transition=False, timeout=0.05)
        return super().did_mount()
    

class PathSettingsView(SettingsView):
    '''Template View for Path Setters'''

    def __init__(self, page: ft.Page):
        super().__init__(page)
        self.dialog = self.build_dialog()

    def build_dialog(self):
        return ft.Banner(
            bgcolor=ft.colors.AMBER_100,
            leading=ft.Icon(ft.icons.WARNING_AMBER_ROUNDED, color=ft.colors.AMBER, size=GUI_PARAMS['icon_size']),
            content=ft.Text(f'Restart Ki-nTree to load the new {self.title}', weight=ft.FontWeight.BOLD),
            actions=[
                ft.TextButton('Discard', on_click=lambda _: self.show_dialog(open=False)),
            ],
        )
    
    def show_dialog(self, d_type=None, message=None, snackbar=False, open=True):
        return super().show_dialog(d_type, message, snackbar, open)


class UserSettingsView(PathSettingsView):
    '''User settings view'''

    title = 'User Settings'
    route = '/settings/user'
    settings = {
        **global_settings.USER_SETTINGS,
        **{
            'DATASHEET_SAVE_ENABLED': global_settings.DATASHEET_SAVE_ENABLED,
            'DATASHEET_SAVE_PATH': global_settings.DATASHEET_SAVE_PATH,
            'AUTOMATIC_BROWSER_OPEN': global_settings.AUTOMATIC_BROWSER_OPEN
        },
        **{
            'CACHE_ENABLED': global_settings.CACHE_ENABLED,
            'CACHE_VALID_DAYS': global_settings.CACHE_VALID_DAYS
        },
    }
    settings_file_list = [
        global_settings.USER_CONFIG_FILE,
        global_settings.CONFIG_GENERAL_PATH,
        global_settings.CONFIG_SEARCH_API_PATH,
    ]

    def save(self):
        # Save all settings
        for sf in self.settings_file_list:
            super().save(settings_file=sf, show_dialog=True)
    
    def increment_cache_value(self, inc):
        field = SETTINGS[self.title]['CACHE_VALID_DAYS'][1]
        current_value = int(field.value)
        if not inc:
            if current_value > 1:
                field.value = f'{current_value - 1}'
        else:
            if current_value < 99:
                field.value = f'{current_value + 1}'
        field.on_change(_=None)
        field.update()

    def build_column(self):
        # Header
        self.column = self.init_column()
        # Fields
        for name, field in self.fields.items():
            self.update_field(name, field, self.column)
    
        # Create refs
        datasheet_row_ref = ft.Ref[ft.Row]()
        cache_row_ref = ft.Ref[ft.Row]()

        # Create row for cache validity
        SETTINGS[self.title]['CACHE_VALID_DAYS'][1].value = self.settings['CACHE_VALID_DAYS']
        cache_row = ft.Row(
            ref=cache_row_ref,
            controls=[
                ft.Text('Keep Cache Valid For (Days): '),
                ft.IconButton(ft.icons.REMOVE, on_click=lambda _: self.increment_cache_value(False)),
                SETTINGS[self.title]['CACHE_VALID_DAYS'][1],
                ft.IconButton(ft.icons.ADD, on_click=lambda _: self.increment_cache_value(True)),
            ],
        )
        self.column.controls.append(cache_row)
        # Add cache row to switch refs
        SETTINGS[self.title]['Enable Supplier Search Cache'][1].refs = [cache_row_ref]

        for name, field in SETTINGS[self.title].items():
            if field[0] in ['AUTOMATIC_BROWSER_OPEN', 'DATASHEET_SAVE_ENABLED', 'DATASHEET_SAVE_PATH', 'DATASHEET_INVENTREE_ENABLED']:
                self.fields[name].on_change = lambda _: self.save()
            elif field[0] in ['CACHE_ENABLED', 'CACHE_VALID_DAYS']:
                self.fields[name].on_change = lambda _: self.save()
        self.settings_file = self.settings_file_list[0]

        # Update datasheet ref
        for idx, field in enumerate(self.column.controls):
            if isinstance(field, SwitchWithRefs):
                if field.label == 'Save Datasheets to Local Folder':
                    datasheet_row_ref.current = self.column.controls[idx + 1]
                    SETTINGS[self.title]['Save Datasheets to Local Folder'][1].refs = [datasheet_row_ref]
        
        # Save button
        self.add_buttons(self.column, test=False)

    def did_mount(self):
        try:
            # Reset Index
            self.navigation_rail.selected_index = 0
            self.navigation_rail.update()
        except AssertionError:
            pass
        return super().did_mount()


class SupplierSettingsView(SettingsView):
    '''Supplier settings view'''

    title = 'Supplier Settings'
    route = '/settings/supplier'
    settings = global_settings.CONFIG_SUPPLIERS
    settings_file = global_settings.CONFIG_SUPPLIERS_PATH

    def __init__(self, page: ft.Page):
        super().__init__(page)

    def save_s(self, e: ft.ControlEvent, supplier: str, show_dialog=True):
        '''Save supplier settings'''

        # Enable/Name settings
        supplier_settings = self.settings
        enable_name = {
            'enable': SETTINGS[self.title][supplier]['Enable'][1].value,
            'name': SETTINGS[self.title][supplier]['InvenTree Name'][1].value,
        }
        supplier_settings.update({supplier: enable_name})
        config_interface.dump_file(supplier_settings, self.settings_file)
        # Update suppliers
        global_settings.load_suppliers()
        
        # API settings
        if supplier == 'Digi-Key':
            from ...search import digikey_api
            # Load settings from file
            settings_from_file = config_interface.load_file(global_settings.CONFIG_DIGIKEY_API)
            # Update settings values
            updated_settings = {
                'DIGIKEY_CLIENT_ID': SETTINGS[self.title][supplier]['Client ID'][1].value,
                'DIGIKEY_CLIENT_SECRET': SETTINGS[self.title][supplier]['Client Secret'][1].value,
            }
            digikey_settings = {**settings_from_file, **updated_settings}
            config_interface.dump_file(digikey_settings, global_settings.CONFIG_DIGIKEY_API)
            digikey_api.setup_environment(force=True)
        elif supplier == 'Mouser':
            from ...search import mouser_api
            # Load settings from file
            settings_from_file = config_interface.load_file(global_settings.CONFIG_MOUSER_API)
            # Update settings values
            updated_settings = {
                'MOUSER_PART_API_KEY': SETTINGS[self.title][supplier]['Part API Key'][1].value,
            }
            mouser_settings = {**settings_from_file, **updated_settings}
            config_interface.dump_file(mouser_settings, global_settings.CONFIG_MOUSER_API)
            mouser_api.setup_environment(force=True)
        elif supplier == 'Element14' or supplier == 'Farnell' or supplier == 'Newark':
            # Load settings from file
            settings_from_file = config_interface.load_file(global_settings.CONFIG_ELEMENT14_API)
            # Update settings values
            updated_settings = {
                'ELEMENT14_PRODUCT_SEARCH_API_KEY': SETTINGS[self.title][supplier]['Product Search API Key (Element14)'][1].value,
                f'{supplier.upper()}_STORE': SETTINGS[self.title][supplier][f'{supplier} Store'][1].value,
            }
            element14_settings = {**settings_from_file, **updated_settings}
            config_interface.dump_file(element14_settings, global_settings.CONFIG_ELEMENT14_API)
        elif supplier == 'LCSC':
            # Load settings from file
            settings_from_file = config_interface.load_file(global_settings.CONFIG_LCSC_API)
            # Update settings values
            updated_settings = {
                'LCSC_API_URL': SETTINGS[self.title][supplier]['API URL'][1].value,
            }
            lcsc_settings = {**settings_from_file, **updated_settings}
            config_interface.dump_file(lcsc_settings, global_settings.CONFIG_LCSC_API)
        elif supplier == 'TME':
            # Load settings from file
            settings_from_file = config_interface.load_file(global_settings.CONFIG_TME_API)
            # Update settings values
            updated_settings = {
                'TME_API_TOKEN': SETTINGS[self.title][supplier]['API Token'][1].value,
                'TME_API_SECRET': SETTINGS[self.title][supplier]['API Secret'][1].value,
                'TME_API_COUNTRY': SETTINGS[self.title][supplier]['API Country'][1].value,
                'TME_API_LANGUAGE': SETTINGS[self.title][supplier]['API Language'][1].value,
            }
            tme_settings = {**settings_from_file, **updated_settings}
            config_interface.dump_file(tme_settings, global_settings.CONFIG_TME_API)
            
        if show_dialog:
            self.show_dialog(
                d_type=DialogType.VALID,
                message=f'{supplier} Settings successfully saved',
            )

    def test_s(self, e: ft.ControlEvent, supplier: str):
        '''Test supplier API settings'''
        self.save_s(e, supplier, show_dialog=False)

        result = False
        if supplier == 'Digi-Key':
            from ...search import digikey_api
            result = digikey_api.test_api()
        elif supplier == 'Mouser':
            from ...search import mouser_api
            result = mouser_api.test_api()
        elif supplier == 'Element14' or supplier == 'Farnell' or supplier == 'Newark':
            from ...search import element14_api
            result = element14_api.test_api()
        elif supplier == 'LCSC':
            from ...search import lcsc_api
            result = lcsc_api.test_api()
        elif supplier == 'TME':
            from ...search import tme_api
            result = tme_api.test_api()

        if result:
            self.show_dialog(
                d_type=DialogType.VALID,
                message=f'Successfully connected to {supplier} API'
            )
        else:
            self.show_dialog(
                d_type=DialogType.ERROR,
                message=f'ERROR: Failed to connect to {supplier} API. Verify the {supplier} credentials and re-try'
            )

    def build_column(self):
        # Header
        self.column = self.init_column()
        
        # Tabs
        supplier_tabs = ft.Tabs(
            selected_index=0,
            animation_duration=10,
            expand=1,
            tabs=[],
        )
        for supplier, settings in SETTINGS[self.title].items():
            supplier_tab_content = [
                ft.Row(height=10),
            ]
            for setting_name, setting_data in settings.items():
                setting_data[1].label = setting_name
                setting_data[1].width = GUI_PARAMS['textfield_width']
                setting_data[1].dense = GUI_PARAMS['textfield_dense']
                setting_data[1].value = setting_data[0]
                supplier_tab_content.extend(
                    [
                        ft.Row([setting_data[1]]),
                        ft.Row(height=GUI_PARAMS['textfield_space_after']),
                    ]
                )

            # Test and Save buttons
            supplier_tab_content.append(
                ft.Row(
                    controls=[
                        ft.ElevatedButton(
                            'Test',
                            width=GUI_PARAMS['button_width'],
                            height=GUI_PARAMS['button_height'],
                            icon=ft.icons.CHECK_OUTLINED,
                            on_click=lambda e, s=supplier: self.test_s(e, supplier=s),
                        ),
                        ft.ElevatedButton(
                            'Save',
                            width=GUI_PARAMS['button_width'],
                            height=GUI_PARAMS['button_height'],
                            icon=ft.icons.SAVE_OUTLINED,
                            on_click=lambda e, s=supplier: self.save_s(e, supplier=s),
                        ),
                    ]
                )
            )

            supplier_tabs.tabs.append(
                ft.Tab(
                    tab_content=ft.Text(supplier, size=16),
                    content=ft.Container(
                        ft.Column(
                            controls=supplier_tab_content,
                        )
                    )
                )
            )

        self.column.controls.append(supplier_tabs)


class InvenTreeSettingsView(SettingsView):
    '''InvenTree settings view'''

    title = 'InvenTree Settings'
    route = '/settings/inventree'
    settings_file = [
        global_settings.INVENTREE_CONFIG,
        global_settings.CONFIG_IPN_PATH,
    ]

    def save(self, file=None, dialog=True):
        address = SETTINGS[self.title]['Server Address'][1].value
        proxy = SETTINGS[self.title]['Proxy'][1].value
        enable_proxy = SETTINGS[self.title]['Enable Proxy Support'][1].value
        if not enable_proxy:
            proxies = None
        elif address.startswith('https'):
            proxies = {'https': proxy}
        else:
            proxies = {'http': proxy}
        if file is None:
            # Save to InvenTree file
            config_interface.save_inventree_user_settings(
                enable=global_settings.ENABLE_INVENTREE,
                server=address,
                username=SETTINGS[self.title]['Username'][1].value,
                password=SETTINGS[self.title]['Password or Token'][1].value,
                enable_proxy=enable_proxy,
                proxies=proxies,
                datasheet_upload=SETTINGS[self.title][
                    'Upload Datasheets to InvenTree'][1].value,
                pricing_upload=SETTINGS[self.title][
                    'Upload Pricing Data to InvenTree'][1].value,
                user_config_path=self.settings_file[0]
            )
            # Alert user
            if dialog:
                self.show_dialog(
                    d_type=DialogType.VALID,
                    message=f'{self.title} successfully saved',
                )
        else:
            super().save(settings_file=file, show_dialog=dialog)

        # Reload InvenTree settings
        global_settings.load_inventree_settings()
        # Reload IPN settings
        global_settings.load_ipn_settings()

    def test(self):
        from ...database import inventree_interface
        self.save(dialog=False)
        connection = inventree_interface.connect_to_server()
        if connection:
            self.show_dialog(
                d_type=DialogType.VALID,
                message='Sucessfully connected to InvenTree server',
            )
        else:
            self.show_dialog(
                d_type=DialogType.ERROR,
                message='Failed to connect to InvenTree server. Check InvenTree credentials are correct and server is running',
            )

    def __init__(self, page: ft.Page):
        # Load InvenTree and IPN settings
        self.settings = {
            **config_interface.load_inventree_user_settings(self.settings_file[0]),
            **config_interface.load_file(self.settings_file[1]),
        }
        super().__init__(page)

    def build_column(self):
        ipn_file = self.settings_file[1]
        ipn_fields = [
            'Default Part Revision',
            'Enable Internal Part Number (IPN)',
            'Use Manufacturer Part Number as IPN',
            'IPN: Enable Prefix',
            'IPN: Prefix',
            'IPN: Enable Category Codes',
            'IPN: Length of Unique ID',
            'IPN: Enable Suffix',
            'IPN: Suffix',
        ]

        # Tabs
        inventree_tabs = ft.Tabs(
            selected_index=0,
            animation_duration=10,
            expand=1,
            tabs=[],
        )
        
        # Build server tab content
        server_col = ft.Column([ft.Row(height=10)])
        for name, field in self.fields.items():
            if name not in ipn_fields:
                self.update_field(name, field, server_col)
        self.add_buttons(server_col, test=True)

        # Add InvenTree server tab
        inventree_tabs.tabs.append(
            ft.Tab(
                tab_content=ft.Text('Server', size=16),
                content=ft.Container(
                    server_col,
                )
            )
        )

        # Link Proxy Switch to the input field
        ref = ft.Ref[ft.TextField]()
        ref.current = SETTINGS[self.title]['Proxy'][1]
        SETTINGS[self.title]['Enable Proxy Support'][1].refs = [ref]

        # Create IPN fields
        ipn_fields_ref = ft.Ref[ft.Row]()
        ipn_fields_col = ft.Column(
            ref=ipn_fields_ref,
            controls=[],
        )
        for name in ipn_fields:
            SETTINGS[self.title][name][1].label = name
            SETTINGS[self.title][name][1].on_change = lambda _: self.save(
                file=ipn_file,
                dialog=False,
            )
            if name.startswith('IPN: '):
                ipn_fields_col.controls.append(
                    ft.Row([SETTINGS[self.title][name][1]])
                )
        ipn_manufacturer_part_number_ref = ft.Ref[ft.Row]()
        ipn_manufacturer_part_number_col = ft.Column(
            ref=ipn_manufacturer_part_number_ref,
            controls=[
                ft.Row([SETTINGS[self.title]['Use Manufacturer Part Number as IPN'][1]]),
                ft.Row([ipn_fields_col]),
            ],
        )
        
        # Build IPN tab column
        ipn_tab_col = ft.Column(
            [
                ft.Row(height=10),
                ft.Row([SETTINGS[self.title]['Default Part Revision'][1]]),
                ft.Row([SETTINGS[self.title]['Enable Internal Part Number (IPN)'][1]]),
                ft.Row([ipn_manufacturer_part_number_col]),
            ]
        )
    
        # Link main IPN switch to corresponding fields
        main_control = 'Enable Internal Part Number (IPN)'
        secondary_control = 'Use Manufacturer Part Number as IPN'
        SETTINGS[self.title][main_control][1].refs = [ipn_manufacturer_part_number_ref]
        SETTINGS[self.title][main_control][1].on_change = lambda _: self.save(
            file=ipn_file,
            dialog=False,
        )

        # Link Manufacturer Part Number switch to corresponding fields
        SETTINGS[self.title][secondary_control][1].refs = [ipn_fields_ref]
        SETTINGS[self.title][secondary_control][1].on_change = lambda _: self.save(
            file=ipn_file,
            dialog=False,
        )

        # Link prefix/suffix switches to corresponding fields
        for name in ['IPN: Enable Prefix', 'IPN: Enable Suffix']:
            ref = ft.Ref[ft.TextField]()
            ref.current = SETTINGS[self.title][name.replace('Enable ', '')][1]
            SETTINGS[self.title][name][1].refs = [ref]

        # Add IPN tab
        inventree_tabs.tabs.append(
            ft.Tab(
                tab_content=ft.Text('Internal Part Number', size=16),
                content=ft.Container(
                    ipn_tab_col,
                )
            )
        )

        # Build column
        self.column = self.init_column()
        # Add tabs
        self.column.controls.append(inventree_tabs)


class KiCadSettingsView(PathSettingsView):
    '''KiCad settings view'''

    title = 'KiCad Settings'
    route = '/settings/kicad'
    settings = global_settings.KICAD_SETTINGS
    settings_file = global_settings.KICAD_CONFIG_PATHS
